/*
 * Copyright(C) 2010 Neil Jagdish Patel
 * Copyright(C) 2010 Canonical Ltd.
 *
 * This library is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License
 * version 3.0 as published by the Free Software Foundation.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License version 3.0 for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library. If not, see
 * <http://www.gnu.org/licenses/>.
 *
 * Authored by Neil Jagdish Patel <neil.patel@canonical.com>
 * Authored by Ken VaDine <ken.vandine@canonical.com>
 */

[DBus (name = "com.Gwibber.Streams")]
private interface StreamsInterface : Object {
        public abstract string Messages (string stream, string account, int time, string transient, string recipient, string orderby, string order, int limit) throws DBus.Error;
}

namespace Gwibber
{
    public class Streams : Object
    {
        private const string streams_name  = "com.Gwibber.Streams";
        private const string streams_path  = "/com/gwibber/Streams";
        private const string streams_iface = "com.Gwibber.Streams";

        private StreamsInterface streams_service;
	private Gwibber.Utils utils;
	private Gwibber.Messages messages_service;
        private Dee.Model? model;
	private uint model_refresh_id;

	/** 
            Streams::is_available:
            @arg0: The current state

            Emitted when com.Gwibber.Streams availability state changes
	*/
	public signal void is_available(bool is_up);

	public Streams ()
	{
            try
            {
                var dbus_conn = DBus.Bus.get(DBus.BusType.SESSION);

                this.streams_service = (StreamsInterface) dbus_conn.get_object(this.streams_name,
                                                               this.streams_path,
                                                               this.streams_iface);                                                               
		this.utils = new Gwibber.Utils();
		this.utils.setup(this.streams_name);
		this.utils.available.connect(this.streams_available);
                messages_service = new Gwibber.Messages ();
                this.model_refresh_id = 0;
            }
            catch (GLib.Error e)
            {
                warning ("Unable to get Gwibber Streams");
            }
        }


        /**
         * com.Gwibber.Streams
         **/
        public string messages(string stream, string account, int time)
        {
            try
            {
                return this.streams_service.Messages(stream, account, time, "0", "0", "time", "DESC", 1000);

            }
            catch (DBus.Error e)
            {
                warning (e.message);
                return "";
            }
          
        }   

	public Dee.Model? get_stream_model()
	{
            this.model = new Dee.SequenceModel();
            this.model.set_schema ("s", "s", "s", "s", "s", "s", "s", "s", "s", "s");
            messages_service.message.connect((change, data) => {
                this.on_refresh(change, data);});
            Idle.add (() => {
                refresh_model ();
                return false;
            });

            return this.model;
            /*
            Dee.Filter *sort_by_time_filter = new Dee.Filter.collator (4);
            var sort_time_model = new Dee.FilterModel (sort_by_time_filter, this.model);
            return sort_time_model;
            */
        }

        private void on_refresh(string change, string data)
        {
            debug ("on_refresh");
            if (this.model_refresh_id != 0)
            {
                return;
            }

            this.model_refresh_id = Timeout.add_seconds (10, () => {
                refresh_model ();
                this.model_refresh_id = 0;
                return false;
            });
            return;
        }

        private void refresh_model()
        {
            debug ("refresh_model");
            this.model.clear();
            var streams = new List<string> ();
            streams.append("messages");
            streams.append("replies");
            streams.append("private");
            streams.append("public");
            streams.append("images");
            streams.append("links");
            streams.append("videos");
            foreach (string stream in streams)
            {
              debug ("STREAM is %s", stream);
              string messages = this.messages(stream, "all", 0);
              var parser = new Json.Parser();
              parser.load_from_data(messages, -1);
              Json.Array nodeArray = parser.get_root().get_array();
              for(int i = 0; i < nodeArray.get_length(); i++) {
                string _account = nodeArray.get_element(i).get_object().get_string_member("account");
                string _stream = nodeArray.get_element(i).get_object().get_string_member("stream");
                string _service = nodeArray.get_element(i).get_object().get_string_member("service");
                Json.Object sender_obj = nodeArray.get_element(i).get_object().get_object_member("sender");
                string sender = sender_obj.get_string_member("name");
                uint _time = (uint)nodeArray.get_element(i).get_object().get_double_member("time");
                if (_time < 1)
                {
                    _time = (uint)nodeArray.get_element(i).get_object().get_int_member("time");
                }
                string text = nodeArray.get_element(i).get_object().get_string_member("text");
                string icon = this.utils.get_avatar_path(sender_obj.get_string_member("image"));
                string url = nodeArray.get_element(i).get_object().get_string_member("url");
                string content = nodeArray.get_element(i).get_object().get_string_member("content");
                string t = this.utils.generate_time_string(_time);
                this.model.append(_account, _stream, _service, sender, _time.to_string(), text, icon, url, content, t);
                //debug ("%s - %s - %s - %s - %s - %s - %s - %s - %s - %s", _account, _stream, _service, sender, _time, text, icon, url, content, t);
              }
            }
            debug ("Model has %u rows", this.model.get_n_rows());
	}

	public void streams_available(bool is_up)
	{
		this.is_available(is_up);
	}
    }
}
